#!/usr/bin/env python3
"""
Utility file that provides various functions.
"""


# pylint: disable=broad-exception-caught
# pylint: disable=protected-access
# pylint: disable=invalid-name
# pylint: disable=import-outside-toplevel
# pylint: disable=import-error
# ruff: noqa: E501


import os
import sys
import traceback
import shlex
from functools import wraps, partial
from subprocess import STDOUT, CalledProcessError, TimeoutExpired, check_output
from time import sleep, time
from types import FunctionType
from typing import TypeVar
from qecore.logger import Logging

# Satisfy the typing hint but without the need to import entire class.
gi_repository_atspi_accessible = TypeVar(
    "gi_repository_atspi_accessible"
)
Application = TypeVar("Application")

QE_DEVELOPMENT = not os.path.isdir("/mnt/tests/")


logging_class = Logging()
LOGGING = logging_class.logger


def non_critical_execution(function):
    """
    Decorator for non critical functions.
    Do no throw the entire execution away because of an error in non critical parts.

    :param function: Function to execute in try catch block.
    :type function: function
    """

    @wraps(function)
    def wrapper_over_function_with_arguments(self, *args, **kwargs) -> None:
        """
        Wrapper for a function accepting arbitrary arguments.
        """

        try:
            return function(self, *args, **kwargs)
        except Exception as error:
            if (
                hasattr(self, "context")
                and self.context
                and hasattr(self.context, "embed")
            ):
                self.context.embed(
                    mime_type="text/plain",
                    data=repr(error),
                    caption=f"Error while executing: '{function.__name__}'",
                )

            # No matter what happens, send the error to stdout.
            LOGGING.info(
                f"_non_critical_execution in function '{function.__name__}'\n{error}"
            )
            traceback.print_exc(file=sys.stdout)

    return wrapper_over_function_with_arguments


def get_application(context, application) -> Application:
    """
    Get Application class instance of an application, based upon given name.

    :type context: <behave.runner.Context>
    :param context: Context object that is passed from common_steps.

    :type application: str
    :param application: String of application identification: name.

    :rtype: <qecore.application.Application>
    :return: Application class instance

    .. note::

        Do **NOT** call this by yourself.
        This function is called by :mod:`common_steps`.
    """

    app_class_to_return = None
    try:
        # Get app from environment file.
        app_class_to_return = getattr(context, application)
    except AttributeError:
        # Get app from sandbox application list.
        for app in context.sandbox.applications:
            if app.component == application:
                app_class_to_return = app
                break
    except TypeError:
        app_class_to_return = context.sandbox.default_application
        assert context.sandbox.default_application is not None, "\n".join(
            (
                "Default application was not found. Check your environment file!",
                "This is indication that no application was defined in environment.py.",
            )
        )

    assert app_class_to_return is not None, "\n".join(
        (
            f"Application '{application}' was not found.",
            "Possibly wrongly defined application in environment.py file",
            "or incorrect use of decorator in .feature file.",
        )
    )

    assert not isinstance(app_class_to_return, str), " ".join((
        "Application class was not found.",
        "Usually indication of not installed application."
    ))

    return app_class_to_return


def get_application_root(  # pylint: disable=unused-argument
    context, application
) -> gi_repository_atspi_accessible:
    """
    Get Accessibility object of an application, based upon given name.

    :type context: <behave.runner.Context>
    :param context: Context object that is passed from common steps.

    :type application: str
    :param application: String of application identification: name.

    :rtype: <dogtail.tree.root.application>
    :return: Return root object of application.

    .. note::

        Do **NOT** call this by yourself.
        This function is called by :mod:`common_steps`.
    """

    from dogtail.tree import (
        root,
    )

    try:
        root_to_return = root.application(application)
    except Exception as error:
        traceback.print_exc(file=sys.stdout)
        raise UserWarning(
            "".join(
                (
                    "Application was not found in accessibility. ",
                    "Check your environment or feature file!",
                )
            )
        ) from error

    return root_to_return


def run(command, timeout=None) -> str:
    """
    Execute a command and get output.

    :type command: str
    :param command: Command to be executed.

    :rtype: str
    :return: Return string value of command output or exception error output.
    """

    try:
        output = check_output(
            command, shell=True, stderr=STDOUT, encoding="utf-8", timeout=timeout
        )
        return output
    except CalledProcessError as error:
        return error.output


def run_verbose(command, timeout=None) -> tuple:
    """
    Execute a command and get output in verbose format.

    :type command: str
    :param command: Command to be executed.

    :rtype: list
    :return: Return list with following format (output, return code, exception).
    """

    try:
        output = check_output(
            command, shell=True, stderr=STDOUT, encoding="utf-8", timeout=timeout
        )
        return (output, 0, None)
    except CalledProcessError as error:
        return (error.output, error.returncode, error)


def run_embed(context, command, *a, **kw):
    """
    Execute a command and get output.
    This version of command run will also embed results to the html log.

    :type context: <behave.runner.Context>
    :param context: Context object that will be used embedding.

    :type command: str
    :param command: Command to be executed.

    :rtype: list
    :return: Return list with following format (output, return code, exception).
    """

    try:
        output = check_output(
            command, shell=True, stderr=STDOUT, *a, **kw, encoding="utf-8"
        )
        return_code = 0
        error = None
    except CalledProcessError as process_error:
        output = process_error.output
        return_code = process_error.returncode
        error = process_error

    context.embed("text/plain", f"$?={return_code}", caption=f"{command} result")
    context.embed("text/plain", output, caption=f"{command} output")

    return (error.output, error.returncode, error)


def overview_action(action="hide") -> None:
    """
    Hide or show application overview.

    :type action: str
    :param action: Hide or show application overview.

    This function takes only 'hide' and 'show' value.
    """

    # DBus calls are deprecated
    # https://gitlab.gnome.org/GNOME/gnome-shell/-/merge_requests/1970

    assert action in ["hide", "show"], "".join(
        (
            "Unknown option, only defined ones are 'show' or 'hide'.",
            f"You tried to use: '{action}'",
        )
    )

    from dogtail.tree import (
        root,
    )

    gnome_shell = root.application("gnome-shell")

    activities_buttons = gnome_shell.findChildren(
        lambda x: x.name == "Activities"
        and (x.roleName == "label" or x.roleName == "toggle button")
    )

    if len(activities_buttons) == 1:
        from dogtail.predicate import (
            GenericPredicate,
        )

        # Get the only child as an object.
        activities_button = activities_buttons[0]

        # Find Ancestor unfortunately takes only Predicates not lambdas alone, \
        # will use it since implementing it here would be too long.

        activities_toggle_button = None
        # If the label was found, find an ancestor with toggle button.
        if activities_button.roleName == "label":
            activities_toggle_button = activities_button.findAncestor(
                GenericPredicate(roleName="toggle button")
            )
        # If label was not found the toggle button is the actual node we found.
        else:
            activities_toggle_button = activities_button

        # Make sure the activities is in wanted state.
        for _ in range(5):
            if (action == "hide" and activities_toggle_button.checked) or (
                action == "show" and not activities_toggle_button.checked
            ):
                activities_toggle_button.click()
                sleep(1)

            else:
                return

    else:
        pass  # possibly in gnome-classic


# behave-common-steps leftover.
def wait_until(
    tested, element=None, timeout=30, period=0.25, params_in_list=False
) -> bool:
    """
    This function keeps running lambda with specified params until the
    result is True or timeout is reached. Instead of lambda, Boolean variable
    can be used instead.

    Sample usages::

        >>> wait_until(lambda x: x.name != 'Loading...', context.app.instance)
        Pause until window title is not 'Loading...'.
        Return False if window title is still 'Loading...'
        Throw an exception if window doesn't exist after default timeout

        >>> wait_until(lambda element, expected: x.text == expected,
            (element, 'Expected text'), params_in_list=True)
        Wait until element text becomes the expected (passed to the lambda)

        >>> wait_until(dialog.dead)
        Wait until the dialog is dead
    """

    if isinstance(tested, bool):

        def curried_func():
            return tested

    # Or if callable(tested) and element is a list or a tuple.
    elif (
        isinstance(tested, FunctionType)
        and isinstance(element, (tuple, list))
        and params_in_list
    ):
        curried_func = partial(tested, *element)
    # Or if callable(tested) and element is not None?
    elif isinstance(tested, FunctionType) and element is not None:
        curried_func = partial(tested, element)
    else:
        curried_func = tested

    exception_thrown = None
    must_end = int(time()) + timeout
    while int(time()) < must_end:
        try:
            if curried_func():
                return True
        except Exception as error:
            # If lambda has thrown the exception we'll re-raise it later
            # and forget about if lambda passes.
            exception_thrown = error
        sleep(period)
    if exception_thrown is not None:
        raise exception_thrown

    return False


KEY_VALUE = {
    "Alt": 64,
    "Alt_L": 64,
    "Alt_R": 108,
    "Shift": 50,
    "Shift_L": 50,
    "Shift_R": 62,
    "Ctrl": 37,
    "Tab": 23,
    "Super": 133,
}


class HoldKey:
    """
    Simple utility to press a key and do some other actions.

    This is a context manager so the usage is as follows::

        >>> with HoldKey("Alt"):
        >>>     do_some_stuff()

    """

    def __init__(self, key_name) -> None:
        # Import only when required.
        from dogtail.rawinput import (
            holdKey,
        )

        self.key_name = key_name
        holdKey(self.key_name)

    def __enter__(self):
        return self

    def __exit__(self, my_type, value, trace) -> None:
        # Import only when required.
        from dogtail.rawinput import (
            releaseKey,
        )

        releaseKey(self.key_name)


class Tuple(tuple):
    """
    Simple tuple class with some defined arithmetic operations.

    :type command: tuple
    :param command: Tuple.
    """

    def __add__(self, other):
        return Tuple(x + y for x, y in zip(self, other))

    def __rmul__(self, other):
        return Tuple(other * x for x in self)

    def __eq__(self, other) -> bool:
        return all(x == y for x, y in zip(self, other))

    def __lt__(self, other) -> bool:
        return self[0] < other[0] or self[1] < other[1]

    def __gt__(self, other) -> bool:
        return self[0] > other[0] or self[1] > other[1]


def get_center(node) -> Tuple:
    """
    Simple utility to get the center of the node.

    :type node: <dogtail.tree.Node>
    :param node: Node passed to the function.

    :rtype: Tuple
    :return: Tuple with coordinates of the center of a Node.
    """

    return Tuple(
        int(node.position[0] + node.size[0] / 2),
        int(node.position[1] + node.size[1] / 2)
    )


def validate_command(command) -> str:
    """
    Simple utility to try and escape any character that is not alpha character.

    :type command: str
    :param command: Command to be executed.

    :rtype: str
    :return: Validated command.
    """

    parsed_command = shlex.split(command)
    valid_command = ""
    for command_part in parsed_command:
        for character in command_part:
            valid_command += f"\\{character}" if not character.isalpha() else character
        valid_command += " "
    return valid_command


def verify_path(template_path):
    """
    Simple utility to verify if the file exists.

    :type template_path: str
    :param template_path: File location.

    :rtype: str or None
    :return: File path if found, None if not found.
    """

    try:
        assert os.path.isfile(template_path)
    except Exception as error:
        raise UserWarning(f"Desired file was not found: {error}") from error
    return template_path




def plain_dump(node) -> None:
    """
    Simple attempt to get more information from a11y tree.

    :type node: <dogtail.tree.root.application>
    :param node: Accessibility node.
    """

    spacer = " "

    def crawl(node, depth) -> None:
        dump(node, depth)
        for child in node.children:
            crawl(child, depth + 1)

    def dump_std_out(item, depth) -> None:
        print("".join(
            (
                spacer * depth,
                str(item),
                f"     [p:{item.position}, ",
                f"s:{item.size}, ",
                f"vis:{item.visible}, ",
                f"show:{item.showing}]"
            )
        ))

    dump = dump_std_out
    crawl(node, 0)


def patch_scenario(scenario, max_attempts=3, stability=False):
    """
    Changing implementation of behave patch to enable stability logic.
    https://github.com/behave/behave/blob/main/behave/contrib/scenario_autoretry.py
    """

    import functools
    from behave.model import ScenarioOutline

    def scenario_run_with_retries(scenario_run, *args, **kwargs):
        final_attempt = 0
        for attempt in range(1, max_attempts + 1):
            final_attempt = attempt
            # Log what attempt are we on.
            if attempt < max_attempts + 1:
                logic = "Stability" if stability else "Auto retry"
                LOGGING.debug(f"{logic} scenario attempt '{attempt}'.")

            # The scenario_run makes the actual scenario run
            # Returns True on failure.
            scenario_run_failed = scenario_run(*args, **kwargs)

            # So if the result is False it did not fail.
            if not scenario_run_failed and not stability:
                # Return False which means it did not fail.
                LOGGING.debug(f"Auto retry scenario passed after '{attempt}' attempts.")
                return False

            # If the run failed and we are testing stability, end the run here as fail.
            if scenario_run_failed and stability:
                # Return True which means it did fail.
                LOGGING.debug(f"Scenario is not stable, fail on attempt '{attempt}'.")
                return True

        # If stability is in effect and we get here, the logic passes.
        if stability:
            # If it got to this point nothing has failed.
            LOGGING.debug(f"The scenario is stable after '{final_attempt}' attempts.")
            return False

        # If stability is not in effect it should end much earlier.
        LOGGING.debug(f"Auto retry scenario failed after '{final_attempt}' attempts.")
        return True

    if isinstance(scenario, ScenarioOutline):
        scenario_outline = scenario
        for scenario in scenario_outline.scenarios:
            scenario_run = scenario.run
            scenario.run = functools.partial(scenario_run_with_retries, scenario_run)
    else:
        scenario_run = scenario.run
        scenario.run = functools.partial(scenario_run_with_retries, scenario_run)


def get_func_params_and_values():
    """
    Simple way to have logging data from the place it was called in.
    """

    # Get callers frame parameters.
    _local_func_parameters = sys._getframe().f_back.f_locals.keys()
    # Get callers frame parameter values.
    _local_func_parameters_value = sys._getframe().f_back.f_locals.values()

    _map_of_keys_and_values = [
        x if "self" == x else x if "context" == x else str(x) + "=" + str(y)
        for x, y in zip(_local_func_parameters, _local_func_parameters_value)
    ]

    return f"({', '.join(_map_of_keys_and_values)})"


class RunAndLog:
    """
    Utility class for better representation and logging of executed command.
    """

    def __init__(self, command, timeout=None):
        self._command = command
        self._timeout = timeout

        self._output = None
        self._return_code = None
        self._error = None

        self.run_and_log()


    def formatted_output(self):
        """
        Based on number of new lines decide how to format it.
        """

        # Lets 'repr' the output if there is only one line.
        # If there are more lines lets keep the log more readable and output it as is.
        if self._output and self._output.count("\n") > 1:
            return self._output

        return repr(self._output)


    def run_and_log(self):
        """
        Run and Log the command execution.
        """

        try:
            LOGGING.debug(f"Executing: '{self._command}'")
            self._output = check_output(
                self._command,
                shell=True,
                stderr=STDOUT,
                encoding="utf-8",
                timeout=self._timeout
            )

            self._return_code = 0
            self._error = None

            LOGGING.debug(
                " ".join((
                    f"Command run: output='{self.formatted_output()}',",
                    f"return_code='{self._return_code}',",
                    f"error='{self._error}'"
                ))
            )

        except CalledProcessError as error:
            self._output = error.output
            self._return_code = error.returncode
            self._error = error
            LOGGING.debug(
                " ".join((
                    f"CalledProcessError caught: output='{self.formatted_output()}',",
                    f"return_code='{self._return_code}',",
                    f"error='{self._error}'"
                ))
            )
        except TimeoutExpired as error:
            self._output = error.output
            self._return_code = -1
            self._error = error
            LOGGING.debug(
                " ".join((
                    f"TimeoutExpired caught: output='{self.formatted_output()}',",
                    f"return_code='{self._return_code}',",
                    f"error='{self._error}'"
                ))
            )

        except Exception as error:
            self._output = error
            self._return_code = -1
            self._error = error
            LOGGING.debug(
                " ".join((
                    f"Unexpected Exception caught: output='{self.formatted_output()}',",
                    f"return_code='{self._return_code}',",
                    f"error='{self._error}'"
                ))
            )


    @property
    def command(self):
        """
        Get and log the command value.
        """

        LOGGING.debug(f"Command value: '{self._command}'")

        return self._command


    @property
    def timeout(self):
        """
        Get and log the timeout value.
        """

        LOGGING.debug(f"Timeout value: '{self._timeout}'")

        return self._timeout


    @property
    def output(self):
        """
        Get and log the output value.
        """

        LOGGING.debug(f"Output value: '{self.formatted_output()}'")

        return self._output


    @property
    def return_code(self):
        """
        Get and log the return_code value.
        """

        LOGGING.debug(f"Return code value: '{self._return_code}'")

        return self._return_code


    @property
    def error(self):
        """
        Get and log the error value.
        """

        LOGGING.debug(f"Error value: '{self._error}'")

        return self._error


    def __str__(self):
        return str(self._output)


def log_message_to_journal(priority, identifier, invoke):
    """
    Invoke a program to be logged in journal.

    :param priority: Priority of the message.
    :type priority: str

    :param identifier: Identifier we want in the journal.
    :type identifier: str

    :param invoke: Invoke a program to execute.
    :type invoke: str
    """

    LOGGING.debug(get_func_params_and_values())

    # journal entry -p priority -t identifier [invoke a program]
    RunAndLog(
        " ".join(
            (
                "systemd-cat",
                f"-p {priority}",
                f"-t {identifier}",
                invoke,
            )
        )
    )



#!/usr/bin/env python3
"""
Icons for xorg/wayland recognition in html-report.
"""


class QECoreIcon:
    """
    QECoreIcon helper class.
    """

    def __init__(self, mime_type, encoding, data) -> None:
        self.mime_type = mime_type
        self.encoding = encoding
        self.data = data

    def to_src(self) -> str:
        """
        Convert present data to icon format.

        :return: Icon data structure in base64 format.
        :rtype: str
        """

        return f"data:{self.mime_type};{self.encoding},{self.data}"


qecore_icons = {
    "x11": QECoreIcon(
        "image/svg+xml",
        "base64",
        """
        PCEtLSBDcmVhdGVkIHdpdGggSW5rc2NhcGUgKGh0dHA6Ly93d3cuaW5rc2NhcGUub3JnLykgLS0+
        Cgo8c3ZnCiAgIHhtbG5zOnN2Zz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciCiAgIHhtbG5z
        PSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyIKICAgdmVyc2lvbj0iMS4xIgogICB3aWR0aD0i
        Mjc1IgogICBoZWlnaHQ9IjI3NSIKICAgaWQ9InN2ZzIiPgogIDxkZWZzCiAgICAgaWQ9ImRlZnM0
        IiAvPgogIDxnCiAgICAgdHJhbnNmb3JtPSJ0cmFuc2xhdGUoLTI5NS4xNzg1NywtMzM4LjYxMjE3
        KSIKICAgICBpZD0ibGF5ZXIxIj4KICAgIDxwYXRoCiAgICAgICBkPSJtIDI5Ni43ODU3MSw2MTIu
        NjMwMDMgYyAwLDAgMTA0LjA2Mjg1LC0xMzIuNzg1OSAxMDQuMDYyODUsLTEzMi43ODU5IDAsMCAt
        MTA0LjA2Mjg1LC0xNDAuMjQ5ODIgLTEwNC4wNjI4NSwtMTQwLjI0OTgyIGwgNjcuMzAyOTksMCBj
        IDAsMCA4My4zNDc0OSwxMTMuNDg4MjYgODMuMzQ3NDksMTEzLjQ4ODI2IGwgLTEyNi44NjQ3Nywx
        NTkuNTQ3NDYgLTIzLjc4NTcxLDAgeiBNIDQxNy4yODc5NCw1MDAuMDY1NzggNTQ0LjExNjYyLDMz
        OS41OTQzMSBjIDAsMCAyNC40NTQ4MSwwIDI0LjQ1NDgxLDAgTCA0NjQuNDc3NDksNDcxLjI2NTc4
        IGMgMCwwIDEwNC4wOTM5NCwxNDEuMzY0MjUgMTA0LjA5Mzk0LDE0MS4zNjQyNSBsIC02Ny4zMDM1
        NywwIGMgMCwwIC04My45Nzk5MiwtMTEyLjU2NDI1IC04My45Nzk5MiwtMTEyLjU2NDI1IHoiCiAg
        ICAgICBpZD0icGF0aDI4MzAiCiAgICAgICBzdHlsZT0iZmlsbDojMDAwMDAwIiAvPgogIDwvZz4K
        PC9zdmc+Cg==
        """.replace(
            "\n", ""
        ),
    ),
    "wayland": QECoreIcon(
        "image/svg+xml",
        "base64",
        """
        PCEtLSBDb3B5cmlnaHQgwqkgMjAxMCBLcmlzdGlhbiBIw7hnc2JlcmcKCiAgICAgUGVybWlzc2lv
        biB0byB1c2UsIGNvcHksIG1vZGlmeSwgZGlzdHJpYnV0ZSwgYW5kIHNlbGwgdGhpcwogICAgIHNv
        ZnR3YXJlIGFuZCBpdHMgZG9jdW1lbnRhdGlvbiBmb3IgYW55IHB1cnBvc2UgaXMgaGVyZWJ5IGdy
        YW50ZWQKICAgICB3aXRob3V0IGZlZSwgcHJvdmlkZWQgdGhhdCB0aGUgYWJvdmUgY29weXJpZ2h0
        IG5vdGljZSBhcHBlYXIgaW4KICAgICBhbGwgY29waWVzIGFuZCB0aGF0IGJvdGggdGhhdCBjb3B5
        cmlnaHQgbm90aWNlIGFuZCB0aGlzIHBlcm1pc3Npb24KICAgICBub3RpY2UgYXBwZWFyIGluIHN1
        cHBvcnRpbmcgZG9jdW1lbnRhdGlvbiwgYW5kIHRoYXQgdGhlIG5hbWUgb2YKICAgICB0aGUgY29w
        eXJpZ2h0IGhvbGRlcnMgbm90IGJlIHVzZWQgaW4gYWR2ZXJ0aXNpbmcgb3IgcHVibGljaXR5CiAg
        ICAgcGVydGFpbmluZyB0byBkaXN0cmlidXRpb24gb2YgdGhlIHNvZnR3YXJlIHdpdGhvdXQgc3Bl
        Y2lmaWMsCiAgICAgd3JpdHRlbiBwcmlvciBwZXJtaXNzaW9uLiAgVGhlIGNvcHlyaWdodCBob2xk
        ZXJzIG1ha2Ugbm8KICAgICByZXByZXNlbnRhdGlvbnMgYWJvdXQgdGhlIHN1aXRhYmlsaXR5IG9m
        IHRoaXMgc29mdHdhcmUgZm9yIGFueQogICAgIHB1cnBvc2UuICBJdCBpcyBwcm92aWRlZCAiYXMg
        aXMiIHdpdGhvdXQgZXhwcmVzcyBvciBpbXBsaWVkCiAgICAgd2FycmFudHkuCgogICAgIFRIRSBD
        T1BZUklHSFQgSE9MREVSUyBESVNDTEFJTSBBTEwgV0FSUkFOVElFUyBXSVRIIFJFR0FSRCBUTyBU
        SElTCiAgICAgU09GVFdBUkUsIElOQ0xVRElORyBBTEwgSU1QTElFRCBXQVJSQU5USUVTIE9GIE1F
        UkNIQU5UQUJJTElUWSBBTkQKICAgICBGSVRORVNTLCBJTiBOTyBFVkVOVCBTSEFMTCBUSEUgQ09Q
        WVJJR0hUIEhPTERFUlMgQkUgTElBQkxFIEZPUiBBTlkKICAgICBTUEVDSUFMLCBJTkRJUkVDVCBP
        UiBDT05TRVFVRU5USUFMIERBTUFHRVMgT1IgQU5ZIERBTUFHRVMKICAgICBXSEFUU09FVkVSIFJF
        U1VMVElORyBGUk9NIExPU1MgT0YgVVNFLCBEQVRBIE9SIFBST0ZJVFMsIFdIRVRIRVIgSU4KICAg
        ICBBTiBBQ1RJT04gT0YgQ09OVFJBQ1QsIE5FR0xJR0VOQ0UgT1IgT1RIRVIgVE9SVElPVVMgQUNU
        SU9OLAogICAgIEFSSVNJTkcgT1VUIE9GIE9SIElOIENPTk5FQ1RJT04gV0lUSCBUSEUgVVNFIE9S
        IFBFUkZPUk1BTkNFIE9GCiAgICAgVEhJUyBTT0ZUV0FSRS4gLS0+Cgo8IS0tIENyZWF0ZWQgd2l0
        aCBJbmtzY2FwZSAoaHR0cDovL3d3dy5pbmtzY2FwZS5vcmcvKSAtLT4KCjxzdmcKICAgeG1sbnM6
        ZGM9Imh0dHA6Ly9wdXJsLm9yZy9kYy9lbGVtZW50cy8xLjEvIgogICB4bWxuczpjYz0iaHR0cDov
        L2NyZWF0aXZlY29tbW9ucy5vcmcvbnMjIgogICB4bWxuczpyZGY9Imh0dHA6Ly93d3cudzMub3Jn
        LzE5OTkvMDIvMjItcmRmLXN5bnRheC1ucyMiCiAgIHhtbG5zOnN2Zz0iaHR0cDovL3d3dy53My5v
        cmcvMjAwMC9zdmciCiAgIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyIKICAgeG1s
        bnM6c29kaXBvZGk9Imh0dHA6Ly9zb2RpcG9kaS5zb3VyY2Vmb3JnZS5uZXQvRFREL3NvZGlwb2Rp
        LTAuZHRkIgogICB4bWxuczppbmtzY2FwZT0iaHR0cDovL3d3dy5pbmtzY2FwZS5vcmcvbmFtZXNw
        YWNlcy9pbmtzY2FwZSIKICAgd2lkdGg9IjI2NiIKICAgaGVpZ2h0PSIyNjYiCiAgIGlkPSJzdmcy
        IgogICB2ZXJzaW9uPSIxLjEiCiAgIGlua3NjYXBlOnZlcnNpb249IjAuNDgrZGV2ZWwgciIKICAg
        c29kaXBvZGk6ZG9jbmFtZT0id2F5bGFuZC5zdmciCiAgIGlua3NjYXBlOmV4cG9ydC1maWxlbmFt
        ZT0iL2hvbWUva3JoL3dheWxhbmQucG5nIgogICBpbmtzY2FwZTpleHBvcnQteGRwaT0iNDEuMzAy
        MzAzIgogICBpbmtzY2FwZTpleHBvcnQteWRwaT0iNDEuMzAyMzAzIj4KICA8ZGVmcwogICAgIGlk
        PSJkZWZzNCIgLz4KICA8c29kaXBvZGk6bmFtZWR2aWV3CiAgICAgaWQ9ImJhc2UiCiAgICAgcGFn
        ZWNvbG9yPSIjZmZmZmZmIgogICAgIGJvcmRlcmNvbG9yPSIjNjY2NjY2IgogICAgIGJvcmRlcm9w
        YWNpdHk9IjEuMCIKICAgICBpbmtzY2FwZTpwYWdlb3BhY2l0eT0iMCIKICAgICBpbmtzY2FwZTpw
        YWdlc2hhZG93PSIyIgogICAgIGlua3NjYXBlOnpvb209IjEuMzQ5MzQ1MyIKICAgICBpbmtzY2Fw
        ZTpjeD0iNDYuOTE4NzUzIgogICAgIGlua3NjYXBlOmN5PSIxNTMuMzE0NjYiCiAgICAgaW5rc2Nh
        cGU6ZG9jdW1lbnQtdW5pdHM9InB4IgogICAgIGlua3NjYXBlOmN1cnJlbnQtbGF5ZXI9ImxheWVy
        MSIKICAgICBzaG93Z3JpZD0iZmFsc2UiCiAgICAgaW5rc2NhcGU6d2luZG93LXdpZHRoPSIxMjc4
        IgogICAgIGlua3NjYXBlOndpbmRvdy1oZWlnaHQ9IjkyNSIKICAgICBpbmtzY2FwZTp3aW5kb3ct
        eD0iMCIKICAgICBpbmtzY2FwZTp3aW5kb3cteT0iMjQiCiAgICAgaW5rc2NhcGU6d2luZG93LW1h
        eGltaXplZD0iMCIKICAgICBmaXQtbWFyZ2luLXRvcD0iMTAiCiAgICAgZml0LW1hcmdpbi1sZWZ0
        PSIxMCIKICAgICBmaXQtbWFyZ2luLXJpZ2h0PSIxMCIKICAgICBmaXQtbWFyZ2luLWJvdHRvbT0i
        MTAiCiAgICAgdW5pdHM9InB4IiAvPgogIDxtZXRhZGF0YQogICAgIGlkPSJtZXRhZGF0YTciPgog
        ICAgPHJkZjpSREY+CiAgICAgIDxjYzpXb3JrCiAgICAgICAgIHJkZjphYm91dD0iIj4KICAgICAg
        ICA8ZGM6Zm9ybWF0PmltYWdlL3N2Zyt4bWw8L2RjOmZvcm1hdD4KICAgICAgICA8ZGM6dHlwZQog
        ICAgICAgICAgIHJkZjpyZXNvdXJjZT0iaHR0cDovL3B1cmwub3JnL2RjL2RjbWl0eXBlL1N0aWxs
        SW1hZ2UiIC8+CiAgICAgICAgPGRjOnRpdGxlIC8+CiAgICAgIDwvY2M6V29yaz4KICAgIDwvcmRm
        OlJERj4KICA8L21ldGFkYXRhPgogIDxnCiAgICAgaW5rc2NhcGU6bGFiZWw9IkxheWVyIDEiCiAg
        ICAgaW5rc2NhcGU6Z3JvdXBtb2RlPSJsYXllciIKICAgICBpZD0ibGF5ZXIxIgogICAgIHN0eWxl
        PSJkaXNwbGF5OmlubGluZSIKICAgICB0cmFuc2Zvcm09InRyYW5zbGF0ZSgtMjA4LjU3MTQyLC0z
        MzIuMDc2NTEpIj4KICAgIDxwYXRoCiAgICAgICBzb2RpcG9kaTp0eXBlPSJhcmMiCiAgICAgICBz
        dHlsZT0iZmlsbDojZmZiYzAwO2ZpbGwtb3BhY2l0eToxO3N0cm9rZTpub25lIgogICAgICAgaWQ9
        InBhdGgyOTg1IgogICAgICAgc29kaXBvZGk6Y3g9IjM0MS40Mjg1NiIKICAgICAgIHNvZGlwb2Rp
        OmN5PSI0NjUuMjE5MzMiCiAgICAgICBzb2RpcG9kaTpyeD0iMTIyLjg1NzE0IgogICAgICAgc29k
        aXBvZGk6cnk9IjEyMi44NTcxNCIKICAgICAgIGQ9Im0gNDY0LjI4NTcsNDY1LjIxOTMzIGMgMCw2
        Ny44NTIxMiAtNTUuMDA1MDIsMTIyLjg1NzE0IC0xMjIuODU3MTQsMTIyLjg1NzE0IC02Ny44NTIx
        MywwIC0xMjIuODU3MTQsLTU1LjAwNTAyIC0xMjIuODU3MTQsLTEyMi44NTcxNCAwLC02Ny44NTIx
        MiA1NS4wMDUwMSwtMTIyLjg1NzE0IDEyMi44NTcxNCwtMTIyLjg1NzE0IDY3Ljg1MjEyLDAgMTIy
        Ljg1NzE0LDU1LjAwNTAyIDEyMi44NTcxNCwxMjIuODU3MTQgeiIgLz4KICAgIDxnCiAgICAgICB0
        cmFuc2Zvcm09InNjYWxlKDAuOTE1NTg5NiwxLjA5MjE5MjQpIgogICAgICAgc3R5bGU9ImZvbnQt
        c2l6ZTo0MzcuNDczODc2OTVweDtmb250LXN0eWxlOm5vcm1hbDtmb250LXZhcmlhbnQ6bm9ybWFs
        O2ZvbnQtd2VpZ2h0OmJvbGQ7Zm9udC1zdHJldGNoOm5vcm1hbDtsaW5lLWhlaWdodDoxMjUlO2xl
        dHRlci1zcGFjaW5nOjBweDt3b3JkLXNwYWNpbmc6MHB4O2ZpbGw6I2ZmZmZmZjtmaWxsLW9wYWNp
        dHk6MTtzdHJva2U6bm9uZSIKICAgICAgIGlkPSJ0ZXh0Mjk4NyI+CiAgICAgIDxwYXRoCiAgICAg
        ICAgIGQ9Im0gNDQzLjkxMDQ5LDMxNC41OTU0OSAyLjM0OTcxLDAgLTEuNDk1MjcsMS40OTUyOCAt
        MC44NTQ0NCwwIDAsLTEuNDk1MjggbSAtMTczLjAyNDM0LDMuNjMxMzggMCwxLjcwODg4IC01LjU1
        Mzg3LDYuODM1NTMgLTAuODU0NDQsMCAtMC44NTQ0NCwtNS4zNDAyNiBjIDMuMjc1MzgsLTIuMTM1
        OTEgNS42OTYyOSwtMy4yMDM5NiA3LjI2Mjc1LC0zLjIwNDE1IG0gMjAyLjUwMjU2LDIuMzQ5NzEg
        YyA0Ljg0MTY0LDEuOWUtNCA3LjU0NzM2LDMuMDYxOTQgOC4xMTcxOSw5LjE4NTI0IGwgLTQuMDU4
        NiwyOC44MzczOSAxLjQ5NTI4LDAgMy4yMDQxNSwtMi4xMzYxIDAuODU0NDQsMCAwLDAuNjQwODMg
        YyAtNC40MTQ4MSwyLjcwNTg5IC04LjY4NzAxLDEwLjExMTAzIC0xMi44MTY2MiwyMi4yMTU0NyAt
        MS41NjY2NiwxMC4zOTU4MyAtNC4yMDExOCwyMC4wMDgyOCAtNy45MDM1OCwyOC44MzczOSAtMy4y
        NzU1Myw5LjExNDE0IC00LjkxMzIxLDE3LjUxNjEzIC00LjkxMzAzLDI1LjIwNjAyIGwgLTYuMTk0
        NywxNS4xNjYzMyBjIC0xLjdlLTQsMS4yODE3MiAyLjM0OTU0LDIuNTYzMzggNy4wNDkxNCwzLjg0
        NDk4IC0wLjg1NDYyLDIuOTkwNiAtMi4xMzYyOCw0LjQ4NTg3IC0zLjg0NDk5LDQuNDg1ODIgbCAw
        LDAuODU0NDQgMS40OTUyNywwIDAsMC42NDA4MyBjIC0xLjdlLTQsMS41NjY1MyAtMi40MjEwOCwy
        LjM0OTc3IC03LjI2Mjc1LDIuMzQ5NzEgbCAwLjg1NDQ0LDAuODU0NDQgMCwyLjk5MDU1IC0xLjQ5
        NTI3LDAgYyAwLjk5NjY4LDIuOTkwNTkgMy4xMzI3OSw0LjQ4NTg2IDYuNDA4MzEsNC40ODU4MSBs
        IDAsMS40OTUyOCAtNi40MDgzMSwwIDAsMC44NTQ0NCBjIDIuMTM1OTQsNGUtNSAzLjIwMzk5LDEu
        MjgxNyAzLjIwNDE2LDMuODQ0OTggbCAtOC45NzE2MywyLjk5MDU1IDAuODU0NDQsMC44NTQ0NCAw
        LDAuNjQwODMgYyAtMS42ZS00LDAuOTk2ODggLTEuMDY4MjEsMS40OTUzIC0zLjIwNDE2LDEuNDk1
        MjcgbCAwLDAuODU0NDQgYyAyLjcwNTU3LDNlLTUgNC4wNTg0NCwwLjQ5ODQ2IDQuMDU4NiwxLjQ5
        NTI3IC0zLjg0NTE1LDEuMjgxNjkgLTguOTAwNTgsNy42MTg3OSAtMTUuMTY2MzMsMTkuMDExMzIg
        LTYuNDA4NDUsMTAuMTEwODkgLTExLjk2MjMxLDE1LjE2NjMzIC0xNi42NjE2MSwxNS4xNjYzMyBs
        IC0xLjcwODg4LDAgYyAtMy41NjAyOSwtMi4yNzg1MSAtNi40Nzk2MywtOS44MjYwNyAtOC43NTgw
        MiwtMjIuNjQyNjkgLTkuMTE0MTUsLTE5LjUwOTcgLTEzLjY3MTE2LC0zMC45NzM0NCAtMTMuNjcx
        MDYsLTM0LjM5MTI2IC0xZS00LC04LjgyOTE2IC0yLjkxOTQ0LC0xOC4yMjggLTguNzU4MDIsLTI4
        LjE5NjU2IC0xLjU2NjU3LC00LjQxNDUyIC0yLjYzNDYyLC05Ljk2ODM5IC0zLjIwNDE2LC0xNi42
        NjE2IGwgLTMuODQ0OTgsLTAuODU0NDQgYyAtOS42ODM3NSwxNC42Njc5OSAtMTQuNzM5MTksMjYu
        MTMxNzMgLTE1LjE2NjMzLDM0LjM5MTI2IC0yLjcwNTgsNC44NDE4OSAtNC44NDE5LDEwLjY4MDU3
        IC02LjQwODMxLDE3LjUxNjA0IDUuODM4NjEsMS41NjY1MiA4Ljc1Nzk1LDMuMDYxNzkgOC43NTgw
        Miw0LjQ4NTgyIGwgMCwwLjg1NDQ0IGMgLTQuNTU3MDksLTAuNTY5NTkgLTcuNzYxMjQsLTEuNjM3
        NjQgLTkuNjEyNDYsLTMuMjA0MTYgLTEuOTkzNzYsMy44NDUwMyAtNC42MjgyOSwxNC41MjU1MyAt
        Ny45MDM1OCwzMi4wNDE1NSAtMi41NjMzOCw4LjExNzE5IC01Ljc2NzUzLDEyLjE3NTc4IC05LjYx
        MjQ3LDEyLjE3NTc4IC0xMC44MjI5NSwtMy41NjAxNiAtMTkuODY1NzgsLTkuMzk4ODQgLTI3LjEy
        ODUsLTE3LjUxNjA0IC0xLjEzOTI4LDJlLTUgLTIuOTkwNTcsLTQuMDU4NTcgLTUuNTUzODcsLTEy
        LjE3NTc5IC0xLjU2NjQ5LDNlLTUgLTIuNjM0NTQsLTQuMDU4NTYgLTMuMjA0MTUsLTEyLjE3NTc4
        IC01LjQxMTQ3LC0xMS41MzQ5IC05LjExNDA1LC0yNC43MDc1MyAtMTEuMTA3NzQsLTM5LjUxNzkx
        IGwgMCwtMjIuMjE1NDcgLTAuODU0NDQsMCAtMy44NDQ5OSwwLjg1NDQ0IDAsLTAuODU0NDQgYyAx
        ZS01LC0wLjk5Njc0IDEuMjgxNjcsLTEuNDk1MTYgMy44NDQ5OSwtMS40OTUyNyBsIC0wLjg1NDQ0
        LC0xMi44MTY2MiAwLC0xMC4wMzk2OCAtNi4xOTQ3LC0zMi4wNDE1NCAwLC0xMy42NzEwNiBjIDIu
        NzA1NzMsLTguNTQ0MjMgNy40MDUxNSwtMTIuODE2NDMgMTQuMDk4MjgsLTEyLjgxNjYyIGwgNDQu
        ODU4MTYsMCAwLDEuNDk1MjcgYyAtMy41NjAyMiwwLjE0MjU5IC01Ljk4MTE0LDAuOTI1ODMgLTcu
        MjYyNzUsMi4zNDk3MiAtMi43MDU3NywtMS41NjYzIC01LjM0MDMsLTIuMzQ5NTMgLTcuOTAzNTgs
        LTIuMzQ5NzIgbCAtOC43NTgwMiwwIC00LjkxMzA0LDAuNjQwODMgLTAuODU0NDQsLTAuNjQwODMg
        LTAuNjQwODMsMCAtNi40MDgzMSwwLjY0MDgzIC0zLjIwNDE1LC0wLjY0MDgzIC0yLjM0OTcyLDAu
        NjQwODMgLTIuNTYzMzIsLTAuNjQwODMgYyAxMC4zOTU2OCwxMS4zOTI3MiAxNi41MTkxNywyMS4w
        MDUxOCAxOC4zNzA0OCwyOC44MzczOSBsIDcuMjYyNzUsMCAwLDEuNDk1MjcgLTUuNTUzODYsMCAw
        LDEuNzA4ODkgYyAyLjEzNjA3LDYuNTUwODYgMy43MDI1NCw5LjgyNjIxIDQuNjk5NDIsOS44MjYw
        NyBsIDEuNzA4ODksMCAzLjg0NDk4LC0wLjg1NDQ0IDAuODU0NDQsMC44NTQ0NCAwLjg1NDQ0LC0w
        Ljg1NDQ0IDAuODU0NDQsMC44NTQ0NCAwLDAuNjQwODMgYyAtMi44NDgxNywxLjcwOTAyIC00Ljc3
        MDY2LDQuMDU4NzMgLTUuNzY3NDcsNy4wNDkxNCBsIC0wLjY0MDgzLDAgMCwxLjQ5NTI3IDEuNDk1
        MjcsMCBjIDEuOTkzNjYsMjMuNzgyMDYgNS42OTYyMyw0MS43OTY1MSAxMS4xMDc3Myw1NC4wNDM0
        MSBsIDMuMjA0MTYsMi4xMzYxIDEuNzA4ODgsMCBjIDIuMTM2MDUsOGUtNSA2LjYyMTg2LC04LjMz
        MDcyIDEzLjQ1NzQ1LC0yNC45OTI0MSA0LjQxNDU0LC03LjU0NzQ1IDcuNjE4NjksLTE0LjE2OTM2
        IDkuNjEyNDYsLTE5Ljg2NTc1IGwgLTAuODU0NDQsLTIuMzQ5NzIgYyA0LjQxNDUzLC01LjU1Mzcz
        IDguOTcxNTUsLTguMzMwNjYgMTMuNjcxMDYsLTguMzMwOCBsIDEuNDk1MjcsMCBjIDIuOTkwNDUs
        MS40ZS00IDUuMTI2NTUsMS41NjY2MSA2LjQwODMxLDQuNjk5NDMgMy45ODcyOSw0LjY5OTU1IDcu
        MTkxNDQsMTEuNzQ4NjggOS42MTI0NiwyMS4xNDc0MiBsIDExLjk2MjE4LDIyLjIxNTQ3IGMgMS41
        NjYzNSwxLjU2NjU2IDIuMzQ5NTksMy4zNDY2NCAyLjM0OTcxLDUuMzQwMjYgNC40MTQ0OSw4LjQw
        MjA3IDguNDAxODgsMTMuOTU1OTMgMTEuOTYyMTgsMTYuNjYxNiBsIDIuNTYzMzIsMCBjIDIuOTkw
        NDEsNmUtNSA2Ljk3NzgsLTEwLjM5NTYzIDExLjk2MjE4LC0zMS4xODcxIDAuODU0MjksOWUtNSAz
        LjIwNCwtNS41NTM3NyA3LjA0OTE0LC0xNi42NjE2MSBsIDAsLTAuODU0NDQgLTEzLjQ1NzQ1LC0x
        LjQ5NTI3IC0zLjIwNDE2LDAuODU0NDQgLTQuMDU4NTksMCAwLC0xLjcwODg4IDguMTE3MTksLTAu
        NjQwODMgMTUuODA3MTYsMCBjIDYuMTIzMzMsLTE4LjM3MDM1IDExLjc0ODQsLTMyLjYxMTAzIDE2
        Ljg3NTIxLC00Mi43MjIwNiAxLjcwODcxLC0yMC43OTEyMiA0LjA1ODQyLC0zMS4xODY5MSA3LjA0
        OTE0LC0zMS4xODcxIGwgOS42MTI0NywtMy44NDQ5OSBtIC00NS40OTg5OSwyLjM0OTcxIGMgLTAu
        ODU0NTksMS45OTM4OSAtMi43MDU4OCwyLjk5MDc0IC01LjU1Mzg3LDIuOTkwNTUgbCAwLC0xLjQ5
        NTI3IDUuNTUzODcsLTEuNDk1MjggbSAtODAuNTMxMDgsMy44NDQ5OSAyMS41NzQ2NCwwIDAsMS40
        OTUyNyAtNC45MTMwNCwwLjY0MDgzIC00LjY5OTQzLC0wLjY0MDgzIC00LjA1ODU5LDEuNDk1Mjcg
        LTEuNDk1MjgsLTEuNDk1MjcgLTYuNDA4MywwIDAsLTEuNDk1MjcgbSAtMTEyLjM1OTAyLDIuOTkw
        NTQgMi4zNDk3MiwwIDEuNDk1MjcsMS40OTUyOCAwLDAuODU0NDQgLTQuNjk5NDMsNC40ODU4MSAw
        LDMuODQ0OTkgLTEuNzA4ODgsMCAwLC0yLjM0OTcxIDAuODU0NDQsLTAuNjQwODQgMCwtMC44NTQ0
        NCAtNC4wNTg2LC0zLjg0NDk4IGMgMC45OTY5MSwtMS45OTM1MiAyLjkxOTQsLTIuOTkwMzYgNS43
        Njc0OCwtMi45OTA1NSBtIDE1Ny4wMDM1Nyw5LjE4NTI1IDIuMzQ5NzEsMCAwLDAuNjQwODMgYyAt
        MS4xZS00LDAuOTk3MDIgLTAuNzgzMzUsMS40OTU0NCAtMi4zNDk3MSwxLjQ5NTI3IGwgLTAuODU0
        NDQsMC44NTQ0NCAwLDAuODU0NDQgLTIuMzQ5NzIsLTAuODU0NDQgMCwtMS40OTUyNyAzLjIwNDE2
        LC0xLjQ5NTI3IG0gLTExLjk2MjE4LDguMzMwOCAwLDAuODU0NDQgYyAtMWUtNCwxLjEzOTQyIC0z
        LjQxNzg2LDIuMzQ5ODggLTEwLjI1MzI5LDMuNjMxMzcgbCAwLC0xLjQ5NTI3IDQuNjk5NDIsLTIu
        MTM2MSAwLjg1NDQ0LDAuNjQwODMgMC42NDA4MywwIDQuMDU4NiwtMS40OTUyNyBtIC0xMzkuNDg3
        NTIsNy42ODk5NyAwLDAuNjQwODMgLTEuNzA4ODgsMS40OTUyNyAtMC42NDA4MywtMC42NDA4MyAw
        LC0wLjg1NDQ0IDAuNjQwODMsLTAuNjQwODMgMS43MDg4OCwwIG0gMTE2LjQxNzYxLDAgMCwwLjY0
        MDgzIGMgLTIuOTkwNjIsMi4yNzg2NiAtNy44MzI0NCw0LjM0MzU2IC0xNC41MjU1LDYuMTk0NyBs
        IDAsLTEuNDk1MjcgYyAxLjEzOTE5LDEuNWUtNCAxLjcwODgyLC0wLjc4MzA5IDEuNzA4ODgsLTIu
        MzQ5NzIgbCAxMi44MTY2MiwtMi45OTA1NCBtIDQwLjU4NTk1LDcuNDc2MzYgMC44NTQ0NSwwLjg1
        NDQ0IDAsMC42NDA4MyAtMS40OTUyOCwxLjcwODg4IC0wLjg1NDQ0LDAgLTEuNzA4ODgsLTEuNzA4
        ODggMCwtMC42NDA4MyAzLjIwNDE1LC0wLjg1NDQ0IG0gMjkuNDc4MjIsMC44NTQ0NCAwLjg1NDQ1
        LDAgMCwwLjY0MDgzIC0xLjQ5NTI4LDEuNzA4ODggLTAuODU0NDQsMCAwLC0wLjg1NDQ0IDEuNDk1
        MjcsLTEuNDk1MjcgbSAtMTMuNDU3NDQsMTAuNjgwNTEgMy4yMDQxNSwyLjEzNjExIGMgLTAuNTY5
        NzYsMS41NjY2MSAtMi4xMzYyMywyLjM0OTg1IC00LjY5OTQzLDIuMzQ5NzEgbCAtMS43MDg4OCwt
        Mi4zNDk3MSBjIC0xLjNlLTQsLTAuNzExOSAxLjA2NzkyLC0xLjQyMzkzIDMuMjA0MTYsLTIuMTM2
        MTEgbSAxOS4yMjQ5Miw5LjgyNjA4IDAuNjQwODMsMC44NTQ0NCAwLDAuNjQwODMgLTAuNjQwODMs
        MC44NTQ0NCAtMS43MDg4OCwwIDAsLTAuODU0NDQgMS43MDg4OCwtMS40OTUyNyBtIC0zMC4zMzI2
        Niw3LjY4OTk3IDAsMC42NDA4MyAtNC4wNTg1OSwzLjIwNDE1IC0yLjM0OTcyLDAgMCwtMC44NTQ0
        NCBjIC0xLjFlLTQsLTEuMjgxNTQgMi4xMzU5OSwtMi4yNzgzOSA2LjQwODMxLC0yLjk5MDU0IG0g
        LTEzMi40MzgzOCwyOS42OTE4MyAwLjg1NDQ0LDAgMCwwLjY0MDgzIGMgLTIuMTM2MDksNC42OTk1
        MSAtMy45ODczNyw3LjU0NzY1IC01LjU1Mzg3LDguNTQ0NDEgbCAtMS43MDg4OCwwIDAsLTMuMjA0
        MTUgNi40MDgzMSwtNS45ODEwOSBtIDEwNC40NTU0MywzNS42NzI5MiAwLjg1NDQ0LDAgMCwxLjQ5
        NTI3IC02LjQwODMsMy4yMDQxNSAtMC44NTQ0NSwtMC44NTQ0NCAwLC0xLjQ5NTI3IDYuNDA4MzEs
        LTIuMzQ5NzEgbSAtODcuNzkzODMsMTAuMDM5NjggMS40OTUyNywwIDIuNTYzMzMsOC4zMzA4IC0x
        LjcwODg4LDAgLTIuMzQ5NzIsLTcuNjg5OTcgMCwtMC42NDA4MyBtIDg1LjQ0NDEyLDkuODI2MDcg
        MC42NDA4MywwIDIuNTYzMzIsMi4zNDk3MiAtMC44NTQ0NCwwLjY0MDgzIC0wLjg1NDQ0LDAgLTEu
        NDk1MjcsLTIuMTM2MSAwLC0wLjg1NDQ1IgogICAgICAgICBpZD0icGF0aDI5ODYiCiAgICAgICAg
        IGlua3NjYXBlOmNvbm5lY3Rvci1jdXJ2YXR1cmU9IjAiCiAgICAgICAgIHNvZGlwb2RpOm5vZGV0
        eXBlcz0iY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY3NjY2NjY2Nj
        Y2NjY2NjY2NjY2NjY3NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2Nj
        Y2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2Nj
        Y2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2NjY2Nj
        Y2NjY2NjY2NjY2NjY2NjY2NjIiAvPgogICAgPC9nPgogIDwvZz4KPC9zdmc+Cg==
        """.replace(
            "\n", ""
        ),
    ),
}



#!/usr/bin/env python3
"""
This file provides loggers.
"""

# ruff: noqa: E501

import os
import sys
import logging

# Singleton solution provided by https://stackoverflow.com/a/54209647


class Singleton(type):
    """
    Singleton class used as metaclass by :py:class:`logger.Logging`.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Logging(metaclass=Singleton):
    """
    Logging class.
    """

    logger = None
    log_file = "/tmp/qecore_logger.log"

    def __init__(self) -> None:
        """
        Initiating logger class with some basic logging setup.
        """

        self.logger = logging.getLogger("qecore_logger")
        self.logger.setLevel(logging.DEBUG)
        # Disable default handler.
        self.logger.propagate = False

        formatter = logging.Formatter(
            "[%(levelname)s] %(asctime)s: [%(filename)s:%(lineno)d] %(func_name)s: %(message)s"
        )

        # Default umask is 0o022 which turns off permissions for groups and others.
        # We need the logger file to be readable and modifiable by anyone.
        os.umask(0)

        # Setup of file handler.
        # All DEBUG and higher level logs will be going to the file.
        file_handler = logging.FileHandler(self.log_file)
        # This log file should be readable and writable by any user.
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        file_handler.set_name("qecore_file_handler")

        # Setup of console handler.
        # All INFO and higher level logs will be going to the console.
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        console_handler.setFormatter(formatter)
        console_handler.set_name("qecore_console_handler")

        self.logger.addHandler(file_handler)
        self.logger.addHandler(console_handler)

        self.logger.addFilter(FuncFilter())

    def qecore_debug_to_console(self) -> None:
        """
        Set file handler level to DEBUG to get the output to console.
        """

        for handler in self.logger.handlers:
            if handler.get_name() == "qecore_console_handler":
                handler.setLevel(logging.DEBUG)
                break

    def qecore_truncate_the_logger_file(self) -> None:
        """
        If the Logging class is used from multiple files or scripts especially when
        running one script from another, the logger is not found and file is deleted,
        making previous logs disappear.

        Lets just null the file on request. In our case we can Null the file in our
        TestSandbox __init__ as that is perfect place to know we started over.
        """

        # Null the file only if it exists.
        # If it does not, no problem, do nothing.
        if os.path.isfile(self.log_file):
            # Default umask is 0o022 which turns off permissions for groups and others.
            os.umask(0)

            # Create a specific file descriptor for our needs with proper permissions.
            descriptor = os.open(
                path=self.log_file,
                flags=(os.O_CREAT | os.O_TRUNC), # Create if not existing, truncate.
                mode=0o666
            )
            # Close the file descriptor.
            os.close(descriptor)


class FuncFilter(logging.Filter):
    """
    Simple logging Filter to get name of the function the log is called from.
    """

    def filter(self, record):
        # Have to walk the frame a bit back.
        # 1. filter, 2. handle, 3. _log, 4. debug, 5. Original calling function.
        record.func_name = str(
            sys._getframe().f_back.f_back.f_back.f_back.f_back.f_code.co_name
        )
        return True



#!/usr/bin/env python3

########################################################################################
# Experimental Atspi API extension                                                     #
########################################################################################

# pylint: disable=broad-exception-caught
# pylint: disable=import-error
# pylint: disable=import-outside-toplevel
# pylint: disable=no-member
# pylint: disable=wrong-import-position
# pylint: disable=wrong-import-order
# ruff: noqa: E402
# ruff: noqa: E501

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

import warnings
warnings.filterwarnings("ignore", "g_object_unref")

from qecore.logger import Logging
from dogtail.rawinput import click

DESKTOP_COORDINATES = 0
WINDOW_COORDINATES = 1

logging_class = Logging()
LOGGING = logging_class.logger

class APIExtension:
    """
    Extension class to extend dogtail API.
    """

    def click_offset(
        self, mouse_button=1, offset_x=0, offset_y=0, coord_type=DESKTOP_COORDINATES
    ):
        """
        Generates a raw click on coordinates of the object.
        User is able to define what mouse button will be used, modify an offset from
        those coordinates and choose a type of coordinates to use.

        :param mouse_button: Mouse button [1 - left, 2 - middle, 3 - right]
        :type mouse_button: int

        :param offset_x: Offset from node coordinates in X axis by number of pixels.
        :type offset_x: int

        :param offset_y: Offset from node coordinates in Y axis by number of pixels.
        :type offset_y: int

        :param coord_type: DESKTOP or WINDOW coordinate type.
        :type coord_type: int
        """

        # Ponytail window id handling for certain situations.
        window_id = self.window_id
        if (
            self.get_name().lower() in ["quit", "exit"]
            or "close" in self.get_name().lower()
        ):
            window_id = ""

        if coord_type == DESKTOP_COORDINATES:
            _position = self.get_component_iface().get_position(DESKTOP_COORDINATES)
        elif coord_type == WINDOW_COORDINATES:
            _position = self.get_component_iface().get_position(WINDOW_COORDINATES)
        else:
            raise RuntimeError(f"Unsupported coordinate type: '{str(coord_type)}'")

        def out_of_bounds(coordinates):
            return (
                coordinates.x > 10000
                or coordinates.y > 10000
                or (coordinates.y == 0 and coordinates.x == 0)
                or coordinates.x < 0
                or coordinates.y < 0
            )

        _size = self.get_component_iface().get_size()

        LOGGING.info(f"Atspi.Accessible position X: {_position.x}")
        LOGGING.info(f"Atspi.Accessible position Y: {_position.y}")
        LOGGING.info(f"Atspi.Accessible size X: {str(int(_size.x))}")
        LOGGING.info(f"Atspi.Accessible size Y: {str(int(_size.y))}")

        if _position and out_of_bounds(_position):
            LOGGING.info(f"Possibly wrong type of coordinate: {str(coord_type)}")

        center_position_x = _position.x + int(_size.x / 2) + offset_x
        center_position_y = _position.y + int(_size.y / 2) + offset_y

        coord_type_str = (
            "DESKTOP_COORDINATES" if coord_type == 0 else "WINDOW_COORDINATES"
        )

        LOGGING.info(
            " ".join(
                (
                    f"Mouse button {mouse_button}",
                    f"click at ({center_position_x},{center_position_y})",
                    f"with coordinate type {coord_type_str}",
                )
            )
        )

        # Using raw click from dogtail for now.
        click(center_position_x, center_position_y, mouse_button, window_id=window_id)


def extend_dogtail_api():
    """
    Attempt to have linter compliant way to extend API.
    """
    Atspi.Accessible.__bases__ = (APIExtension,) + Atspi.Accessible.__bases__
